import inspect
from typing import Any, Awaitable, Callable, Optional, TypeVar, Union, cast

from deprecation import deprecated

from jetpack import _utils
from jetpack.config import _symbols
from jetpack.core._errors import NotAsyncError
from jetpack.core._jetpack_client import default_jetpack_client
from jetpack.core._jetroutine import Jetroutine
from jetpack.core._jetroutine import schedule as schedule

T = TypeVar("T")
__pdoc__ = {}
__pdoc__["jet"] = "Alias for jetroutine"
__pdoc__["function"] = "Alias for jetroutine"
DecoratedFunction = Callable[..., Awaitable[T]]


# @jetroutine is our general remote work decorator. It does not specify how the
# work will be done (RPC, job, queue, etc) and instead leaves that as an
# implementation detail.
def jetroutine(
    fn: Optional[Callable[..., T]] = None,
    # The stararg forces the arguments after that to be specified with the keyword
    #
    # > Parameters after “*” [...] are keyword-only parameters and may only be passed by keyword arguments
    # https://docs.python.org/3/reference/compound_stmts.html#function-definitions
    *,
    with_checkpointing: bool = False,
    endpoint_path: Optional[str] = None,
) -> Union[Jetroutine[T], Callable[[Callable[..., T]], Jetroutine[T]]]:
    """
    Decorator that wraps any async Python function, and runs it as a distributed function.

    Async python functions decorated with @jetroutine will be registered with
    the runtime when not running locally.
    Calling these functions on Jetpack will run them as remote distributed functions.

    #### Arguments:
    * **fn ([Callable[..., T])**: The function being decorated.
    * **with_checkpointing (bool)**: Determines whether the jetroutine's execution will be re-tried upon failure until it succeeds or exceeds the max number of retries. Also see @workflow for the canonical way user programs should set this.

    #### Returns:
     * **Jetroutine[T]**: A Jetroutine that executes the wrapped function


    #### Raises:
    * **NotAsyncError (jetpack.errors.NotAsyncError)**: Raised if the function being decorated is not async.
    * **ApplicationError (jetpack.errors.ApplicationError)**: Raised if the jetroutine's business logic raises an error. This is usually caused by incorrect business logic. `ApplicationError` is a subtype of `JetpackError` (`jetpack.errors.JetpackError`)
    * **SystemError (jetpack.errors.SystemError)**: Raised if there was a jetpack-system or kubernetes error during execution of the jetroutine. This is usually not an error caused by incorrect business logic. `SystemError` is a subtype of `JetpackError` (`jetpack.errors.JetpackError`)"""

    def wrapper(fn: Callable[..., T]) -> Jetroutine[T]:
        # Use asyncio.iscoroutine() instead?
        if not inspect.iscoroutinefunction(fn):
            raise NotAsyncError(
                f"Jetpack functions must be async. {_utils.qualified_func_name(fn)} is not async."
            )
        _symbols.get_symbol_table().register(fn, endpoint_path)
        jetroutineFunction: Jetroutine[T] = default_jetpack_client.jetroutine(
            fn, with_checkpointing
        )
        return jetroutineFunction

    return wrapper(fn) if fn else wrapper


@deprecated(details="Use jetroutine instead.")
def function(
    fn: Optional[Callable[..., T]] = None, *, with_checkpointing: bool = False
) -> Union[Jetroutine[T], Callable[[Callable[..., T]], Jetroutine[T]]]:
    # can enable this print to be more aggressive, since smarthop and other
    # customers may not have python's dev-mode on to see the @deprecated decorator's
    # warning message:
    # print("WARNING: @function is deprecated. Use @jetroutine instead.")
    """
    .. deprecated:: 0.5.0
       Use @jetroutine instead
    """

    return jetroutine(fn, with_checkpointing=with_checkpointing)


@deprecated(details="Use jetroutine instead.")
def jet(
    fn: Optional[Callable[..., T]] = None, *, with_checkpointing: bool = False
) -> Union[Jetroutine[T], Callable[[Callable[..., T]], Jetroutine[T]]]:
    return jetroutine(fn, with_checkpointing=with_checkpointing)


def workflow(
    fn: Optional[Callable[..., T]] = None
) -> Union[Jetroutine[T], Callable[[Callable[..., T]], Jetroutine[T]]]:
    """
    Workflow is syntactical sugar for @jetroutine(with_checkpointing=true).

    It is introduced as its own decorator because `with_checkpointing=true` changes the runtime semantics of the jetroutine being invoked. This change also applies transitively to any child jetroutines that are invoked by the decorated function.
    #### Arguments:
      * **fn: Optional[Callable[..., T] = None**: The function being decorated. Must be an `async` function
      * **endpoint_path: Optional[str] = None**: Optional endpoint path to publicly expose the function. See `@endpoint` for the canonical way user programs should set this.

    #### Returns:
      * **Jetroutine[T]**: A Jetroutine that executes the wrapped function

    #### Raises:
      * NotAsyncError (jetpack.errors.NotAsyncError): Raised if the function being decorated is not async.
      * ApplicationError (jetpack.errors.ApplicationError): Raised if the jetroutine's business logic raises an error. This is usually caused by incorrect business logic. `ApplicationError` is a subtype of `JetpackError` (`jetpack.errors.JetpackError`)
      * SystemError (jetpack.errors.SystemError): Raised if there was a jetpack-system or kubernetes error during execution of the jetroutine. This is usually not an error caused by incorrect business logic. `SystemError` is a subtype of `JetpackError` (`jetpack.errors.JetpackError`)
    """
    return jetroutine(fn, with_checkpointing=True)


def endpoint(
    fn: Optional[Callable[..., T]] = None,
    path: str = "",
) -> Union[Jetroutine[T], Callable[[Callable[..., T]], Jetroutine[T]]]:
    """endpoint is syntactical sugar for @jetroutine(path="/some/path")."""
    return jetroutine(fn, endpoint_path=path)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import abc
import grpc
import jetpack.proto.runtime.v1alpha1.jetworker_pb2

class JetworkerStub:
    def __init__(self, channel: grpc.Channel) -> None: ...
    StartJetroutine: grpc.UnaryUnaryMultiCallable[
        jetpack.proto.runtime.v1alpha1.jetworker_pb2.StartJetroutineRequest,
        jetpack.proto.runtime.v1alpha1.jetworker_pb2.StartJetroutineResponse]


class JetworkerServicer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def StartJetroutine(self,
        request: jetpack.proto.runtime.v1alpha1.jetworker_pb2.StartJetroutineRequest,
        context: grpc.ServicerContext,
    ) -> jetpack.proto.runtime.v1alpha1.jetworker_pb2.StartJetroutineResponse: ...


def add_JetworkerServicer_to_server(servicer: JetworkerServicer, server: grpc.Server) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: jetpack/proto/runtime/v1alpha1/app.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from jetpack.proto.runtime.v1alpha1 import remote_pb2 as jetpack_dot_proto_dot_runtime_dot_v1alpha1_dot_remote__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(jetpack/proto/runtime/v1alpha1/app.proto\x12\x07userapp\x1a+jetpack/proto/runtime/v1alpha1/remote.proto\"\xd9\x01\n\x07UserApp\x12^\n\x14revision_to_manifest\x18\x01 \x03(\x0b\x32(.userapp.UserApp.RevisionToManifestEntryB\x02\x18\x01R\x12revisionToManifest\x12\'\n\x0flatest_manifest\x18\x02 \x01(\tR\x0elatestManifest\x1a\x45\n\x17RevisionToManifestEntry\x12\x10\n\x03key\x18\x01 \x01(\x04R\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\x87\x01\n\x0b\x63om.userappB\x08\x41ppProtoP\x01Z2go.jetpack.io/proto/jetpack/proto/runtime/v1alpha1\xa2\x02\x03UXX\xaa\x02\x07Userapp\xca\x02\x07Userapp\xe2\x02\x13Userapp\\GPBMetadata\xea\x02\x07Userappb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'jetpack.proto.runtime.v1alpha1.app_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\013com.userappB\010AppProtoP\001Z2go.jetpack.io/proto/jetpack/proto/runtime/v1alpha1\242\002\003UXX\252\002\007Userapp\312\002\007Userapp\342\002\023Userapp\\GPBMetadata\352\002\007Userapp'
  _USERAPP_REVISIONTOMANIFESTENTRY._options = None
  _USERAPP_REVISIONTOMANIFESTENTRY._serialized_options = b'8\001'
  _USERAPP.fields_by_name['revision_to_manifest']._options = None
  _USERAPP.fields_by_name['revision_to_manifest']._serialized_options = b'\030\001'
  _USERAPP._serialized_start=99
  _USERAPP._serialized_end=316
  _USERAPP_REVISIONTOMANIFESTENTRY._serialized_start=247
  _USERAPP_REVISIONTOMANIFESTENTRY._serialized_end=316
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import abc
import grpc
import jetpack.proto.runtime.v1alpha1.remote_pb2

class RemoteExecutorStub:
    def __init__(self, channel: grpc.Channel) -> None: ...
    CreateTask: grpc.UnaryUnaryMultiCallable[
        jetpack.proto.runtime.v1alpha1.remote_pb2.CreateTaskRequest,
        jetpack.proto.runtime.v1alpha1.remote_pb2.CreateTaskResponse]

    CancelTask: grpc.UnaryUnaryMultiCallable[
        jetpack.proto.runtime.v1alpha1.remote_pb2.CancelTaskRequest,
        jetpack.proto.runtime.v1alpha1.remote_pb2.CancelTaskResponse]

    PostResult: grpc.UnaryUnaryMultiCallable[
        jetpack.proto.runtime.v1alpha1.remote_pb2.PostResultRequest,
        jetpack.proto.runtime.v1alpha1.remote_pb2.PostResultResponse]

    WaitForResult: grpc.UnaryUnaryMultiCallable[
        jetpack.proto.runtime.v1alpha1.remote_pb2.WaitForResultRequest,
        jetpack.proto.runtime.v1alpha1.remote_pb2.WaitForResultResponse]

    GetTask: grpc.UnaryUnaryMultiCallable[
        jetpack.proto.runtime.v1alpha1.remote_pb2.GetTaskRequest,
        jetpack.proto.runtime.v1alpha1.remote_pb2.GetTaskResponse]

    GetSymbolFromEndpoint: grpc.UnaryUnaryMultiCallable[
        jetpack.proto.runtime.v1alpha1.remote_pb2.GetSymbolFromEndpointRequest,
        jetpack.proto.runtime.v1alpha1.remote_pb2.GetSymbolFromEndpointResponse]
    """GetSymbolFromEndpoint is not used by SDK. It's a way for runtime web server to find
    symbols of endpoints to handle endpoint http requests.
    """

    RegisterApp: grpc.UnaryUnaryMultiCallable[
        jetpack.proto.runtime.v1alpha1.remote_pb2.RegisterAppRequest,
        jetpack.proto.runtime.v1alpha1.remote_pb2.RegisterAppResponse]

    DeregisterApp: grpc.UnaryUnaryMultiCallable[
        jetpack.proto.runtime.v1alpha1.remote_pb2.DeregisterAppRequest,
        jetpack.proto.runtime.v1alpha1.remote_pb2.DeregisterAppResponse]

    RemoteCall: grpc.UnaryUnaryMultiCallable[
        jetpack.proto.runtime.v1alpha1.remote_pb2.RemoteCallRequest,
        jetpack.proto.runtime.v1alpha1.remote_pb2.RemoteCallResponse]
    """Golang prototype"""


class RemoteExecutorServicer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def CreateTask(self,
        request: jetpack.proto.runtime.v1alpha1.remote_pb2.CreateTaskRequest,
        context: grpc.ServicerContext,
    ) -> jetpack.proto.runtime.v1alpha1.remote_pb2.CreateTaskResponse: ...

    @abc.abstractmethod
    def CancelTask(self,
        request: jetpack.proto.runtime.v1alpha1.remote_pb2.CancelTaskRequest,
        context: grpc.ServicerContext,
    ) -> jetpack.proto.runtime.v1alpha1.remote_pb2.CancelTaskResponse: ...

    @abc.abstractmethod
    def PostResult(self,
        request: jetpack.proto.runtime.v1alpha1.remote_pb2.PostResultRequest,
        context: grpc.ServicerContext,
    ) -> jetpack.proto.runtime.v1alpha1.remote_pb2.PostResultResponse: ...

    @abc.abstractmethod
    def WaitForResult(self,
        request: jetpack.proto.runtime.v1alpha1.remote_pb2.WaitForResultRequest,
        context: grpc.ServicerContext,
    ) -> jetpack.proto.runtime.v1alpha1.remote_pb2.WaitForResultResponse: ...

    @abc.abstractmethod
    def GetTask(self,
        request: jetpack.proto.runtime.v1alpha1.remote_pb2.GetTaskRequest,
        context: grpc.ServicerContext,
    ) -> jetpack.proto.runtime.v1alpha1.remote_pb2.GetTaskResponse: ...

    @abc.abstractmethod
    def GetSymbolFromEndpoint(self,
        request: jetpack.proto.runtime.v1alpha1.remote_pb2.GetSymbolFromEndpointRequest,
        context: grpc.ServicerContext,
    ) -> jetpack.proto.runtime.v1alpha1.remote_pb2.GetSymbolFromEndpointResponse:
        """GetSymbolFromEndpoint is not used by SDK. It's a way for runtime web server to find
        symbols of endpoints to handle endpoint http requests.
        """
        pass

    @abc.abstractmethod
    def RegisterApp(self,
        request: jetpack.proto.runtime.v1alpha1.remote_pb2.RegisterAppRequest,
        context: grpc.ServicerContext,
    ) -> jetpack.proto.runtime.v1alpha1.remote_pb2.RegisterAppResponse: ...

    @abc.abstractmethod
    def DeregisterApp(self,
        request: jetpack.proto.runtime.v1alpha1.remote_pb2.DeregisterAppRequest,
        context: grpc.ServicerContext,
    ) -> jetpack.proto.runtime.v1alpha1.remote_pb2.DeregisterAppResponse: ...

    @abc.abstractmethod
    def RemoteCall(self,
        request: jetpack.proto.runtime.v1alpha1.remote_pb2.RemoteCallRequest,
        context: grpc.ServicerContext,
    ) -> jetpack.proto.runtime.v1alpha1.remote_pb2.RemoteCallResponse:
        """Golang prototype"""
        pass


def add_RemoteExecutorServicer_to_server(servicer: RemoteExecutorServicer, server: grpc.Server) -> None: ...
